#ifndef FARRAY_HPP
#define FARRAY_HPP

#include <fstream.h>
#include <stdlib.h> // size_t


template <class T>
class ptr
{
public:
  ptr(T*);
  ~ptr();
  T& operator*() const;
private:
  ptr(const ptr<T>&);
  ptr<T>& operator=(const ptr<T>&);
  T* p;
};

template <class T> class FileArrayProxy;

template <class T>
class FileArray
{
public:
  FileArray(const char* name, size_t size); // create one.
  FileArray(const char* name); // use existing array
  //  destructor created by compiler
  T operator[](size_t size) const;
  FileArrayProxy<T> operator[](size_t size);
  size_t size() const;
private:
  FileArray(const FileArray<T>&); // illegal
  FileArray<T>& operator=(const FileArray<T>&); // illegal
  T readElement(size_t index) const;
  void storeElement(size_t index, const T& elem);
  ptr<iostream> pstream;
  size_t max_size;

  friend class FileArrayProxy<T>;
};


template <class T> class FileArrayProxy
{
public:
  operator T() const;
  FileArrayProxy<T>& operator=(const T& t);
  FileArrayProxy<T>& operator=(const FileArrayProxy<T>& p);
private:
  FileArrayProxy(FileArray<T>& f, size_t i);

  size_t index;
  FileArray<T>& fa;

  friend class FileArray<T>;
};


template <class T>
ptr<T>::ptr(T* tp)
  : p(tp)
{
}

template <class T>
ptr<T>::~ptr()
{
  delete p;
}

template <class T>
T& ptr<T>::operator*() const
{
  return *p;
}

template <class T>
FileArray<T>::FileArray(const char* name, size_t size)
  : pstream(new fstream(name, ios::in|ios::out|ios::binary)),
    max_size(size)
{
  // what if the file could not be opened?

  // store the size on file.
  (*pstream).write((const char*)&max_size, sizeof(max_size));
  // what if write failed?

  T t;
  storeElement(max_size-1,t);
  // What if this failed?
}

template <class T>
FileArray<T>::FileArray(const char* name)
  : pstream(new fstream(name, ios::in|ios::out|ios::binary)),
    max_size(0)
{
  (*pstream).read((char*)&max_size, sizeof(max_size));
  // what if read fails or max_size == 0?
}

template <class T>
T FileArray<T>::operator[](size_t size) const
{
  // what if size >= max_size?
  return readElement(size);
}

template <class T>
FileArrayProxy<T> FileArray<T>::operator[](size_t size)
{
  // what if size >= max_size?
  return FileArrayProxy<T>(*this , size);
}

template <class T>
size_t FileArray<T>::size() const
{
  return max_size;
}



template <class T>
T FileArray<T>::readElement(size_t index) const
{
  // what if index >= max_elements?
  (*pstream).seekg(sizeof(max_size)+index*sizeof(T), ios::beg);
  // what if seek fails?

  T t;
  (*pstream).read((char*)&t, sizeof(t));
  // what if read fails?

  return t;
}

template <class T>
void FileArray<T>::storeElement(size_t index, const T& elem)
{
  (*pstream).seekp(sizeof(max_size)+index*sizeof(T), ios::beg);
  (*pstream).write((char*)&elem, sizeof(elem));

  // what if wite failed?
}


// template <class T>
// ostream& operator<<(ostream& os, const FileArrayProxy<T>& p)
// {
//   os << p.operator T();
// }


template <class T>
FileArrayProxy<T>::FileArrayProxy(FileArray<T>& f, size_t i)
  : index(i),
    fa(f)
{
}

template <class T>
FileArrayProxy<T>::operator T() const
{
  return fa.readElement(index);
}

template <class T>
FileArrayProxy<T>& FileArrayProxy<T>::operator=(const T& t)
{
  fa.storeElement(index,t);
  return *this;
}

template <class T>
FileArrayProxy<T>& FileArrayProxy<T>::operator=(const FileArrayProxy<T>& p)
{
  fa.storeElement(index,p);
  return *this;
}

#endif // FARRAY_HPP

